# -*- encoding: utf-8 -*-
from mimetypes import guess_type
import time

from django.contrib.contenttypes.forms import generic_inlineformset_factory
from django.core.signing import Signer, BadSignature
from django.core.urlresolvers import reverse
from django.db.models import Q
from django.http import HttpResponseRedirect, HttpResponse, Http404
from django.shortcuts import get_object_or_404
from django.utils.http import urlencode
from django.utils.timezone import now
from django.views.generic import TemplateView, ListView, DetailView, View, CreateView

from helpdesk import Filter
from helpdesk.forms import CommentForm, TicketForm, FilterForm, TicketCreateForm, SearchForm
from helpdesk.models import Ticket, HistoryAction, Comment, MailAttachment
from helpdesk.signals import new_answer, ticket_updated


class TestView(TemplateView):
    template_name = 'test.html'


class HomeView(ListView):
    template_name = 'helpdesk/home.html'
    filter_form = None
    filter = None
    paginate_by = 20
    search = None
    search_form = None

    def _get_list_template(self):
        mode = self.request.session.get('mode', 'normal')
        if mode not in ['normal', 'compact']:
            mode = 'normal'

        if mode == 'normal':
            return 'helpdesk/ticket_list.html'
        else:
            return 'helpdesk/ticket_list_compact.html'

    def get_context_data(self, **kwargs):
        context = super(HomeView, self).get_context_data(**kwargs)
        context['filter_form'] = self.filter_form
        context['list_template'] = self._get_list_template()
        context['search_form'] = self.search_form
        context['search'] = self.search
        context['extra'] = urlencode({'search': self.search}) if self.search else None
        return context

    def dispatch(self, request, *args, **kwargs):
        self.search_form = SearchForm(request.GET)
        if self.search_form.is_valid():
            self.search = self.search_form.cleaned_data.get('search', None)

        self.filter = Filter(request)
        initial = self.filter.get_form_init()
        initial.update(
            {'mode': request.session.get('mode', 'normal')}
        )
        self.filter_form = FilterForm(data=request.POST or None, initial=initial,
                                      email_filter=request.user.has_perm('helpdesk.view_customer'),
                                      view_assignees=request.user.has_perm('helpdesk.view_customer'))
        return super(HomeView, self).dispatch(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        if self.filter_form.is_valid():
            self.filter.by_assignee(self.filter_form.cleaned_data['assignee'])
            self.filter.by_state(self.filter_form.cleaned_data['state'])
            self.filter.by_project(self.filter_form.cleaned_data['project'])
            self.filter.by_email(self.filter_form.cleaned_data.get('email', None))

            self.request.session['mode'] = self.filter_form.cleaned_data['mode']

            return HttpResponseRedirect(reverse('helpdesk_home'))
        return self.render_to_response(self.get_context_data())

    def get_queryset(self):
        queryset = Ticket.objects.all()
        filters = self.filter.get_filters()
        if filters:
            queryset = queryset.filter(**filters)
        if self.search:
            keywords = [w for w in self.search.split(' ') if w]
            print(keywords)
            qs = Q()
            for word in keywords:
                qs |= Q(customer__icontains=word) | Q(title__icontains=word) | Q(body__icontains=word)
            queryset = queryset.filter(qs)
        return queryset.order_by('-updated')


AttachmentFormset = generic_inlineformset_factory(MailAttachment, can_delete=False, extra=3)


class TicketCreateView(CreateView):
    model = Ticket
    template_name = 'helpdesk/ticket_create.html'
    form_class = TicketCreateForm
    attachment_formset = None

    def dispatch(self, request, *args, **kwargs):
        self.attachment_formset = AttachmentFormset(data=request.POST or None, files=request.FILES or None)
        return super(TicketCreateView, self).dispatch(request, *args, **kwargs)

    def form_valid(self, form):
        if not self.attachment_formset.is_valid():
            return self.render_to_response(self.get_context_data())
        data = form.cleaned_data
        comment = data.pop('comment')
        ticket = Ticket.create(body='This ticket was created by [user:%d]' % self.request.user.pk,
                               message_id='ticket-%d' % time.mktime(now().timetuple()),
                               author=self.request.user,
                               **data)
        reply = Comment.objects.create(ticket=ticket, body=comment, author=self.request.user)

        self.attachment_formset.instance = reply
        self.attachment_formset.save()

        new_answer.send(sender=Comment, ticket=ticket, answer=reply)
        return HttpResponseRedirect(reverse('helpdesk_home'))

    def get_context_data(self, **kwargs):
        context = super(TicketCreateView, self).get_context_data(**kwargs)
        context['attachment_formset'] = self.attachment_formset
        return context


class EmailView(View):
    def get(self, request, *args, **kwargs):
        ticket = get_object_or_404(Ticket, pk=kwargs['pk'])
        return HttpResponse(ticket.body)


class CommentEmailView(View):
    def get(self, request, *args, **kwargs):
        comment = get_object_or_404(Comment, pk=kwargs['pk'])
        return HttpResponse(comment.body)


class TicketView(DetailView):
    template_name = 'helpdesk/ticket.html'
    model = Ticket
    reply_form = None
    ticket_form = None
    attachment_formset = None

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        self.reply_form = CommentForm(request.POST or None if 'reply' in request.POST else None)
        self.ticket_form = TicketForm(request.POST or None if 'ticket' in request.POST else None,
                                      instance=self.object)
        self.attachment_formset = AttachmentFormset(data=request.POST or None if 'reply' in request.POST else None,
                                                    files=request.FILES or None)
        return super(TicketView, self).dispatch(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        ticket = self.get_object()
        self.object = ticket
        if 'reply' in request.POST and self.reply_form.is_valid() and self.attachment_formset.is_valid():
            reply = self.reply_form.save(commit=False)
            reply.author = request.user
            reply.ticket = ticket
            reply.save()

            self.attachment_formset.instance = reply
            self.attachment_formset.save()

            ticket.state = self.reply_form.cleaned_data['state']
            ticket.save()

            new_answer.send(sender=Comment, ticket=ticket, answer=reply)
            HistoryAction.objects.create(
                user=reply.author,
                change='added %scomment and set state to <i>%s</i>' % (
                    'internal ' if reply.internal else '',
                    reply.ticket.state),
                ticket=reply.ticket
            )

            return HttpResponseRedirect(reverse('helpdesk_home'))
        elif 'ticket' in request.POST and self.ticket_form.is_valid():
            changes = []
            for fieldname in self.ticket_form.changed_data:
                old_value = self._get_display_value(fieldname)
                setattr(self.object, fieldname, self.ticket_form.cleaned_data[fieldname])
                new_value = self._get_display_value(fieldname)
                changes.append('%s: %s → %s' % (fieldname, old_value, new_value))
            self.object.save(update_fields=self.ticket_form.changed_data)
            if changes:
                HistoryAction.objects.create(
                    user=request.user,
                    change='\n'.join(changes),
                    ticket=self.object
                )
                ticket_updated.send(sender=Ticket,
                                    changed_data=self.ticket_form.changed_data,
                                    changes=changes,
                                    ticket=self.object,
                                    updater=request.user)
            return HttpResponseRedirect(reverse('helpdesk_home'))
        return self.render_to_response(self.get_context_data())

    def _get_display_value(self, fieldname):
        try:
            return getattr(self.object, 'get_%s_display' % fieldname)()
        except:
            return getattr(self.object, fieldname)

    def get_comments(self):
        return self.object.comment_set.all().order_by('created')

    def get_context_data(self, **kwargs):
        context = super(TicketView, self).get_context_data(**kwargs)
        context['comments'] = self.get_comments()
        context['reply_form'] = self.reply_form
        context['ticket_form'] = self.ticket_form
        context['history'] = self.object.historyaction_set.all().order_by('-created')
        context['attachments'] = self.object.attachments.all()
        context['attachment_formset'] = self.attachment_formset
        return context


class AttachmentView(View):
    def dispatch(self, request, *args, **kwargs):
        try:
            mid = Signer().unsign(kwargs['signature'])

            a = get_object_or_404(MailAttachment, pk=mid)
            attachment = a.attachment
            response = HttpResponse()
            response['X-Sendfile'] = attachment.path.encode('utf-8')
            response['Content-Length'] = attachment.size
            response['Content-Type'] = guess_type(attachment.name.split('/')[-1])[0]
            response['Content-Disposition'] = 'attachment; filename=%s' % attachment.name.split('/')[-1]
            return response
        except BadSignature:
            raise Http404



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings
import django.core.files.storage


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Comment',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('body', models.TextField()),
                ('internal', models.BooleanField(help_text='If checked this comment will not be emailed to client', default=False)),
                ('notified', models.BooleanField(editable=False, default=True)),
                ('message_id', models.CharField(max_length=256, blank=True, null=True)),
                ('author', models.ForeignKey(related_name='helpdesk_answers', blank=True, null=True, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='HelpdeskProfile',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('signature', models.TextField(blank=True, null=True)),
                ('send_notifications', models.BooleanField(default=True)),
                ('label', models.CharField(max_length=255, blank=True, null=True)),
                ('user', models.OneToOneField(to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='HistoryAction',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('change', models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name='MailAttachment',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('object_id', models.PositiveIntegerField()),
                ('attachment', models.FileField(storage=django.core.files.storage.FileSystemStorage(base_url='/helpdesk/attachments/', location='/home/ookami/projects/attachments'), upload_to='tickets')),
                ('content_type', models.ForeignKey(to='contenttypes.ContentType')),
            ],
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('machine_name', models.CharField(primary_key=True, serialize=False, max_length=64)),
                ('title', models.CharField(max_length=255)),
                ('email', models.EmailField(max_length=254)),
                ('default_assignee', models.ForeignKey(to=settings.AUTH_USER_MODEL, blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='ProjectAlias',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('email', models.EmailField(unique=True, max_length=254)),
                ('assignee', models.ForeignKey(to=settings.AUTH_USER_MODEL, blank=True, null=True)),
                ('project', models.ForeignKey(to='helpdesk.Project')),
            ],
        ),
        migrations.CreateModel(
            name='State',
            fields=[
                ('machine_name', models.CharField(primary_key=True, serialize=False, max_length=32)),
                ('title', models.CharField(max_length=64)),
                ('resolved', models.BooleanField(default=False)),
                ('color', models.CharField(choices=[('default', 'Default'), ('primary', 'Primary'), ('success', 'Success'), ('warning', 'Warning'), ('danger', 'Danger'), ('info', 'Info')], max_length=32, default='default')),
            ],
        ),
        migrations.CreateModel(
            name='Ticket',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('title', models.CharField(max_length=255)),
                ('body', models.TextField()),
                ('priority', models.IntegerField(choices=[(0, 'Low'), (1, 'Normal'), (2, 'High')], default=1)),
                ('customer', models.EmailField(max_length=254)),
                ('message_id', models.CharField(unique=True, max_length=255)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('assignee', models.ForeignKey(to=settings.AUTH_USER_MODEL, blank=True, null=True)),
                ('project', models.ForeignKey(to='helpdesk.Project', blank=True, null=True)),
                ('state', models.ForeignKey(to='helpdesk.State', default='open')),
            ],
            options={
                'permissions': (('view_customer', 'Can view ticket customer'), ('view_tickets', 'Can view tickets'), ('view_all_tickets', 'Can view tickets assigned to other users')),
            },
        ),
        migrations.AddField(
            model_name='historyaction',
            name='ticket',
            field=models.ForeignKey(to='helpdesk.Ticket'),
        ),
        migrations.AddField(
            model_name='historyaction',
            name='user',
            field=models.ForeignKey(to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='comment',
            name='ticket',
            field=models.ForeignKey(to='helpdesk.Ticket'),
        ),
    ]



# -*- encoding: utf-8 -*-
from django.contrib.auth.models import Group
from django.core.management import BaseCommand
from helpdesk.models import State


class Command(BaseCommand):
    def handle(self, *args, **options):
        State.objects.get_or_create(machine_name='open', color='danger', resolved=False, title='Open')
        State.objects.get_or_create(machine_name='resolved', color='success', resolved=True, title='Resolved')
        Group.objects.get_or_create(name='Helpdesk support')


# -*- encoding: utf-8 -*-
import logging
import re

from django.core.files.base import ContentFile
from django.core.management import BaseCommand
from imbox import Imbox

from helpdesk import SETTINGS
from helpdesk.models import Ticket, Comment, Project, MailAttachment, ProjectAlias

logger = logging.getLogger('helpdesk.mail')


class BlackList(object):
    def __contains__(self, value):
        return value not in ['script']


class Command(BaseCommand):
    server = None
    AUTOREPLY_HEADERS = ['auto-submitted', 'x-autoreply', 'x-autogenerated']

    def _get_initial_issue(self, message):
        subject = getattr(message, 'subject', 'Email ticket')
        customer = message.sent_from[0]['email']
        pattern = re.compile(r'.*\[HD-(\d+)\].*')
        m = pattern.match(subject)
        if m:
            pk = m.group(1)
        else:
            return None

        try:
            return Ticket.objects.get(customer=customer, pk=pk)
        except Ticket.DoesNotExist:
            return None

    def _is_autoreply(self, message):
        for header in message.headers:
            try:
                if header['Name'].lower() in self.AUTOREPLY_HEADERS and header['Value'].lower() != 'no':
                    return True
            except:
                pass
        return False

    def handle_messages(self, imbox, project, email=None, assignee=None):
        logger.info("--- Started processing emails ---")
        unread_messages = imbox.messages(unread=True, folder='INBOX', sent_to=email or project.email)
        for uid, message in unread_messages:
            try:
                subject = getattr(message, 'subject', 'Email ticket')
                logger.info("Got message %s: %s" % (uid, subject))

                if SETTINGS['ignore_autoreply'] and self._is_autoreply(message):
                    logger.info('  Message is auto-submitted, skipping')
                    if SETTINGS['mark_seen']:
                        imbox.mark_seen(uid)
                    continue

                if Ticket.objects.filter(message_id=uid).exists() or Comment.objects.filter(message_id=uid).exists():
                    logger.info('  Message already exists, skipping')
                    if SETTINGS['mark_seen']:
                        imbox.mark_seen(uid)
                    continue

                try:
                    body = message.body['html'][0]
                except IndexError:
                    try:
                        body = message.body['plain'][0]
                    except IndexError:
                        body = ''

                initial = self._get_initial_issue(message)

                if initial is None:
                    ticket = Ticket.create(
                            title=subject[:255],
                            body=body,
                            customer=message.sent_from[0]['email'],
                            message_id=uid,
                            assignee=assignee or project.default_assignee,
                            project=project
                    )
                    logger.info('  Created new ticket')
                    self._create_attachments(message.attachments, ticket)
                else:
                    comment = Comment.objects.create(
                            body=body,
                            author=None,
                            message_id=uid,
                            ticket=initial
                    )
                    self._create_attachments(message.attachments, comment)
                    logger.info('  Added comment to ticket')

                if SETTINGS['mark_seen']:
                    imbox.mark_seen(uid)
            except Exception:
                logger.exception('  Error while retrieving email %s' % uid)
        logger.info("--- Finished processing emails ---")

    def handle(self, *args, **options):
        logger.info("===== Started check_mail =====")
        imbox = Imbox(SETTINGS['imap_host'],
                      username=SETTINGS['username'],
                      password=SETTINGS['password'],
                      ssl=True)

        for project in Project.objects.all():
            logger.info("=== Processing project %s ===" % project.title)
            self.handle_messages(imbox, project)

        for alias in ProjectAlias.objects.all():
            logger.info("=== Processing alias '{alias}' for project '{project}' ===".format(
                    alias=alias.email,
                    project=alias.project.title)
            )
            self.handle_messages(imbox, alias.project, email=alias.email, assignee=alias.assignee)

        imbox.logout()
        logger.info("===== Finished check_mail =====")

    def _create_attachments(self, attachments, obj):
        for attachment in attachments:
            f = ContentFile(attachment['content'].read(), name=attachment.get('filename', 'Attachment'))
            mail_attachment = MailAttachment(
                    attachment=f, content_object=obj
            )
            mail_attachment.save()
